/*
 * Copyright (c) 2010-2011. Axon Framework
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.axonframework.eventsourcing;

import org.axonframework.domain.AggregateIdentifier;
import org.axonframework.domain.AggregateRoot;
import org.axonframework.domain.DomainEvent;
import org.axonframework.domain.DomainEventStream;
import org.axonframework.domain.Event;
import org.axonframework.eventstore.EventStore;
import org.axonframework.eventstore.EventStreamNotFoundException;
import org.axonframework.repository.AggregateNotFoundException;
import org.axonframework.repository.LockingRepository;
import org.axonframework.repository.LockingStrategy;
import org.axonframework.unitofwork.CurrentUnitOfWork;
import org.axonframework.unitofwork.UnitOfWorkListenerAdapter;

import java.util.ArrayList;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import javax.annotation.Resource;

/**
 * Abstract repository implementation that allows easy implementation of an Event Sourcing mechanism. It will
 * automatically publish new events to the given {@link org.axonframework.eventhandling.EventBus} and delegate event
 * storage to the provided {@link org.axonframework.eventstore.EventStore}.
 *
 * @param <T> The type of aggregate this repository stores
 * @author Allard Buijze
 * @see EventSourcedAggregateRoot
 * @see org.axonframework.eventsourcing.AbstractEventSourcedAggregateRoot
 * @see org.axonframework.eventsourcing.annotation.AbstractAnnotatedAggregateRoot
 * @see org.axonframework.eventstore.EventStore
 * @see org.axonframework.eventstore.fs.FileSystemEventStore
 * @since 0.1
 */
public abstract class EventSourcingRepository<T extends EventSourcedAggregateRoot> extends LockingRepository<T>
        implements AggregateFactory<T> {

    private volatile EventStore eventStore;
    private ConflictResolver conflictResolver;
    private Deque<EventStreamDecorator> eventStreamDecorators = new LinkedList<EventStreamDecorator>();

    /**
     * Initializes a repository with the default locking strategy.
     *
     * @see org.axonframework.repository.LockingRepository#LockingRepository()
     */
    protected EventSourcingRepository() {
    }

    /**
     * Initialize a repository with the given locking strategy.
     *
     * @param lockingStrategy the locking strategy to apply to this
     */
    protected EventSourcingRepository(final LockingStrategy lockingStrategy) {
        super(lockingStrategy);
    }

    /**
     * Perform the actual saving of the aggregate. All necessary locks have been verified.
     *
     * @param aggregate the aggregate to store
     */
    @Override
    protected void doSaveWithLock(T aggregate) {
        DomainEventStream eventStream = aggregate.getUncommittedEvents();
        Iterator<EventStreamDecorator> iterator = eventStreamDecorators.descendingIterator();
        while (iterator.hasNext()) {
            eventStream = iterator.next().decorateForAppend(getTypeIdentifier(), aggregate, eventStream);
        }
        eventStore.appendEvents(getTypeIdentifier(), eventStream);
    }

    /**
     * Delegates to {@link #doSaveWithLock(EventSourcedAggregateRoot)}, as Event Sourcing generally doesn't delete
     * aggregates (not their events).
     * <p/>
     * This method may be safely overridden for special cases that do require deleting an Aggregate's Events.
     *
     * @param aggregate the aggregate to delete
     */
    @Override
    protected void doDeleteWithLock(T aggregate) {
        doSaveWithLock(aggregate);
    }

    /**
     * Perform the actual loading of an aggregate. The necessary locks have been obtained.
     *
     * @param aggregateIdentifier the identifier of the aggregate to load
     * @param expectedVersion     The expected version of the loaded aggregate
     * @return the fully initialized aggregate
     *
     * @throws AggregateDeletedException in case an aggregate existed in the past, but has been deleted
     * @throws org.axonframework.repository.AggregateNotFoundException
     *                                   when an aggregate with the given identifier does not exist
     */
    @Override
    protected T doLoad(AggregateIdentifier aggregateIdentifier, final Long expectedVersion) {
        DomainEventStream events;
        try {
            events = eventStore.readEvents(getTypeIdentifier(), aggregateIdentifier);
        } catch (EventStreamNotFoundException e) {
            throw new AggregateNotFoundException(aggregateIdentifier, "The aggregate was not found", e);
        }
        for (EventStreamDecorator decorator : eventStreamDecorators) {
            events = decorator.decorateForRead(getTypeIdentifier(), aggregateIdentifier, events);
        }

        final T aggregate = createAggregate(aggregateIdentifier, events.peek());
        List<DomainEvent> unseenEvents = new ArrayList<DomainEvent>();
        aggregate.initializeState(new CapturingEventStream(events, unseenEvents, expectedVersion));
        CurrentUnitOfWork.get().registerListener(new ConflictResolvingListener(aggregate, unseenEvents));
        return aggregate;
    }

    private List<DomainEvent> asList(DomainEventStream domainEventStream) {
        List<DomainEvent> unseenEvents = new ArrayList<DomainEvent>();
        while (domainEventStream.hasNext()) {
            unseenEvents.add(domainEventStream.next());
        }
        return unseenEvents;
    }

    /**
     * {@inheritDoc}
     * <p/>
     * This implementation is aware of the AggregateSnapshot type events. When <code>firstEvent</code> is an instance
     * of {@link AggregateSnapshot}, the aggregate is extracted from the event. Otherwise, aggregate creation is
     * delegated to the abstract {@link #instantiateAggregate(org.axonframework.domain.AggregateIdentifier,
     * org.axonframework.domain.DomainEvent)} method.
     */
    @Override
    @SuppressWarnings({"unchecked"})
    public T createAggregate(AggregateIdentifier aggregateIdentifier, DomainEvent firstEvent) {
        T aggregate;
        if (AggregateSnapshot.class.isInstance(firstEvent)) {
            aggregate = (T) ((AggregateSnapshot) firstEvent).getAggregate();
        } else {
            aggregate = instantiateAggregate(aggregateIdentifier, firstEvent);
        }
        return aggregate;
    }

    /**
     * Instantiate the aggregate using the given aggregate identifier and first event. The first event of the event
     * stream is passed to allow the repository to identify the actual implementation type of the aggregate to create.
     * The first event can be either the event that created the aggregate or, when using event sourcing, a snapshot
     * event. In either case, the event should be designed, such that these events contain enough information to deduct
     * the actual aggregate type.
     * <p/>
     * Note that aggregate state should <strong>*not*</strong> be initialized by this method. That means, no events
     * should be applied by a call to this method. The first event is passed to allow the implementation to base the
     * exact type of aggregate to instantiate on that event.
     *
     * @param aggregateIdentifier the aggregate identifier of the aggregate to instantiate
     * @param firstEvent          The first event in the event stream. This is either the event generated during
     *                            creation of the aggregate, or a snapshot event
     * @return an aggregate ready for initialization using a DomainEventStream.
     */
    protected abstract T instantiateAggregate(AggregateIdentifier aggregateIdentifier, DomainEvent firstEvent);

    /**
     * {@inheritDoc}
     * <p/>
     * This implementation will do nothing if a conflict resolver (See {@link #setConflictResolver(ConflictResolver)}
     * is set. Otherwise, it will call <code>super.validateOnLoad(...)</code>.
     */
    @Override
    protected void validateOnLoad(T aggregate, Long expectedVersion) {
        if (conflictResolver == null) {
            super.validateOnLoad(aggregate, expectedVersion);
        }
    }

    /**
     * Sets the event store that would physically store the events.
     *
     * @param eventStore the event bus to publish events to
     */
    @Resource
    public void setEventStore(EventStore eventStore) {
        this.eventStore = eventStore;
    }

    /**
     * Sets the Event Stream Decorators that will process the event in the DomainEventStream when read, or written to
     * the event store.
     * <p/>
     * When appending events to the event store, the processors are invoked in the reverse order, causing the first
     * decorator in this list to receive each event first. When reading from events, the decorators are invoked in the
     * order given.
     *
     * @param eventProcessors The processors to that will process events in the DomainEventStream
     */
    public void setEventStreamDecorators(List<? extends EventStreamDecorator> eventProcessors) {
        this.eventStreamDecorators.addAll(eventProcessors);
    }

    /**
     * Sets the snapshotter trigger for this repository.
     *
     * @param snapshotterTrigger the snapshotter trigger for this repository.
     */
    public void setSnapshotterTrigger(SnapshotterTrigger snapshotterTrigger) {
        this.eventStreamDecorators.add(snapshotterTrigger);
    }

    /**
     * Sets the conflict resolver to use for this repository. If not set (or <code>null</code>), the repository will
     * throw an exception if any unexpected changes appear in loaded aggregates.
     *
     * @param conflictResolver The conflict resolver to use for this repository
     */
    public void setConflictResolver(ConflictResolver conflictResolver) {
        this.conflictResolver = conflictResolver;
    }

    private final class ConflictResolvingListener extends UnitOfWorkListenerAdapter {

        private final T aggregate;
        private final List<DomainEvent> unseenEvents;

        private ConflictResolvingListener(T aggregate, List<DomainEvent> unseenEvents) {
            this.aggregate = aggregate;
            this.unseenEvents = unseenEvents;
        }

        @Override
        public void onPrepareCommit(Set<AggregateRoot> aggregateRoots, List<Event> events) {
            if (hasPotentialConflicts()) {
                conflictResolver.resolveConflicts(asList(aggregate.getUncommittedEvents()), unseenEvents);
            }
        }

        private boolean hasPotentialConflicts() {
            return aggregate.getUncommittedEventCount() > 0
                    && aggregate.getVersion() != null
                    && !unseenEvents.isEmpty();
        }
    }

    /**
     * Wrapper around a DomainEventStream that captures all passing events of which the sequence number is larger than
     * the expected version number.
     */
    private static final class CapturingEventStream implements DomainEventStream {

        private final DomainEventStream eventStream;
        private final List<DomainEvent> unseenEvents;
        private final Long expectedVersion;

        private CapturingEventStream(DomainEventStream events,
                                     List<DomainEvent> unseenEvents,
                                     Long expectedVersion) {
            eventStream = events;
            this.unseenEvents = unseenEvents;
            this.expectedVersion = expectedVersion;
        }

        @Override
        public boolean hasNext() {
            return eventStream.hasNext();
        }

        @Override
        public DomainEvent next() {
            DomainEvent next = eventStream.next();
            if (expectedVersion != null && next.getSequenceNumber() > expectedVersion) {
                unseenEvents.add(next);
            }
            return next;
        }

        @Override
        public DomainEvent peek() {
            return eventStream.peek();
        }
    }
}
